#!/usr/bin/env python
#
# Output snarf alerts from the DNS pipeline mediator in CEF format
# ------------------------------------------------------------------------
# Copyright (C) 2012 Carnegie Mellon University. All Rights Reserved.
# ------------------------------------------------------------------------
# Authors: Tony Cebzanov <tonyc@cert.org>
# -----------------------------------------------------------------------------
# GNU Public License (GPL) Rights pursuant to Version 2, June 1991
# Government Purpose License Rights (GPLR) pursuant to DFARS 252.227.7013
# -----------------------------------------------------------------------------
#
# Sample Usage:
#
#     snarf2cef.py tcp://127.0.0.1:5556


import sys
import syslog
from snarf import *
from snarf.alert import datetime_from_timestamp
from optparse import OptionParser
import signal

syslog_priorities = [
    syslog.LOG_NOTICE,
    syslog.LOG_WARNING,
    syslog.LOG_ERR,
    syslog.LOG_CRIT,
    syslog.LOG_ALERT
]

# Explicitly handle SIGINT and raise KeyboardInterrupt so that
# it's not handled by another thread
def sigint_handler(signum, frame):
    raise KeyboardInterrupt

def cef_header(alert):
    tags = ','.join(alert.get_tags())
    return "CEF:0|%s|%s|%s|%s|%s|%d|" %("org.cert.netsa",
                                         alert.get_generator(),
                                         alert.get_generator_version(),
                                         tags,
                                         tags,
                                         alert.get_severity())


def timestamp_to_cef(t):
    dt = datetime_from_timestamp(t * 1000)
    return dt.strftime("%b %d %Y %H:%M:%S")

def handle_sink(alert):
    global options
    global syslog_priorities
    flow_start = int(alert.get_fields()['flow.flowStartMilliseconds'][0])
    flow_end = int(alert.get_fields()['flow.flowEndMilliseconds'][0])
    flow_sip = alert.get_fields()['flow.sourceIPv4Address'][0]
    flow_dip = alert.get_fields()['flow.destinationIPv4Address'][0]
    flow_dport = int(alert.get_fields()['flow.destinationTransportPort'][0])
    query_type = int(alert.get_fields()['dns.queryType'][0])
    authoritative = int(alert.get_fields()['dns.authoritative'][0])
    nxdomain = alert.get_fields()['dns.nxdomain'][0]

    header = cef_header(alert)
    fields = "start=%s end=%s src=%s dst=%s dpt=%d cn1=%d cn2=%d cs1=%s" %(
                             timestamp_to_cef(flow_start),
                             timestamp_to_cef(flow_end),
                             flow_sip,
                             flow_dip,
                             flow_dport,
                             query_type,
                             authoritative,
                             nxdomain
    )
    message = "%s%s" %(header, fields)
    if options.syslog:
        priority = syslog_priorities[ int(alert.get_severity()) - 1]
        syslog.syslog(priority, message)
    else:
        print message


def main():
    global options
    parser = OptionParser()
    parser.add_option("-s", "--syslog", dest="syslog", help="use syslog",
                      action="store_true")
    (options, args) = parser.parse_args()

    if len(args) < 1:
        raise Exception("usage: %s socket" %(sys.argv[0]))
    socket=args[0]
    if options.syslog:
        syslog.openlog("snarf", syslog.LOG_PID, syslog.LOG_LOCAL4)

    signal.signal(signal.SIGINT, sigint_handler)

    try:
        sink = Sink(handle_sink, origin=socket,
                    channel="org.cert.netsa.dns-receiver")
        while True:
            sink.get_thread().join(1000)
            if not sink.get_thread().isAlive:
                break
    except KeyboardInterrupt:
        sink.shutdown()


if __name__ == '__main__':
    main()
